package main

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"math"
	"math/rand"
	"os"
	"strings"
	"time"
)

const (
	_ = iota
	GATEWAY
	SENSOR
)

type GridJSON struct {
	Sensors  [][]position `json:"sensors"`
	Gateways [][]position `json:"gateways"`
}

type position struct {
	X, Y int
}

type settings struct {
	width, height                                  int
	numSensors, numGateways                        int
	minDistanceGwGw, minDistanceSensorSensor       int
	maxDistanceSensorGw, minDistanceSensorGw       int
	minGwPerSensor, maxSensorPerGw, minSensorPerGw int
	numGrids                                       int
	filename                                       string
}

func main() {

	rand.Seed(time.Now().UTC().UnixNano())

	s := getUserInput()

	attempts := 0
	maxAttempts := 1000

	g := GridJSON{}
	for times := 0; times < s.numGrids; times++ {

	Top:
		if attempts > maxAttempts {
			fmt.Println("Unable to create a grid that satisfies all criteria. Please try again.")
			os.Exit(0)
		}

		grid := make([][]int, s.height)
		for i := range grid {
			grid[i] = make([]int, s.width)
		}

		gateways, err := generateGatewayPositions(s.width, s.height, s.numGateways, float64(s.minDistanceGwGw))

		if err != nil {
			attempts++
			goto Top
		}

		//g.Gateways = append(g.Gateways, gateways)
		for _, g := range gateways {
			grid[g.Y][g.X] = GATEWAY
		}

		attempts = 0

		sensors, err := generateSensorPositions(s.width, s.height, s.numSensors, s.minGwPerSensor, s.minSensorPerGw, s.maxSensorPerGw, float64(s.minDistanceSensorSensor), float64(s.minDistanceSensorGw), float64(s.maxDistanceSensorGw), gateways)

		if err != nil {
			attempts++
			goto Top
		}

		g.Gateways = append(g.Gateways, gateways)
		g.Sensors = append(g.Sensors, sensors)
		for _, s := range sensors {
			grid[s.Y][s.X] = SENSOR
		}

		fmt.Printf("\no%so\n", strings.Repeat("-", s.width*2))
		for i := range grid {
			var line = "|"
			for j := 0; j < len(grid[i])*2; j += 2 {
				if grid[i][j/2] == GATEWAY {
					line += "G "
				} else if grid[i][j/2] == SENSOR {
					line += "S "
				} else {
					line += "  "
				}
			}
			line += "|"
			fmt.Printf("%s\n", line)
		}
		fmt.Printf("o%so\n", strings.Repeat("-", s.width*2))
	}

	// Save to JSON
	b, _ := json.Marshal(g)
	err := ioutil.WriteFile(s.filename, b, 0644)
	if err != nil {
		panic(err)
	}
}

func generateGatewayPositions(width, height, numberOfGateways int, minimumDistanceBetween float64) ([]position, error) {

	positions := make([]position, 0, numberOfGateways)

	x := rand.Intn(width)
	y := rand.Intn(height)
	p := position{x, y}
	positions = append(positions, p)

	i := 1
	attempts := 0
	maxAttempts := 1000

Top:
	for i < numberOfGateways {

		if attempts > maxAttempts {
			return nil, fmt.Errorf("Too many attempts. Try again.")
		}

		x := rand.Intn(width)
		y := rand.Intn(height)
		p := position{x, y}

		for j := 0; j < len(positions); j++ {
			dist := calculateDistance(p, positions[j])
			if dist < minimumDistanceBetween {
				attempts++
				goto Top
			}
		}

		// We made it!
		i++
		positions = append(positions, p)

	}

	return positions, nil
}

func generateSensorPositions(width, height, numberOfSensors, minGwPerSensor, minSensorsPerGw, maxSensorPerGw int, minDistSensorSensor, minDistToGw, maxDistToGw float64, gateways []position) ([]position, error) {

	positions := make([]position, 0, numberOfSensors)

	sensorsPerGateway := make([]int, len(gateways))

	i := 0
	attempts := 0
	maxAttempts := 1000

Top:
	for i < numberOfSensors {

		if attempts > maxAttempts {
			return nil, fmt.Errorf("Too many attempts. Try again.")
		}

		x := rand.Intn(width)
		y := rand.Intn(height)
		p := position{x, y}

		// Check that sensor is not too close to any other sensors
		for j := 0; j < len(positions); j++ {
			if calculateDistance(p, positions[j]) < minDistSensorSensor {
				attempts++
				goto Top
			}
		}

		// Check that sensor is not too close to any gateways,
		// and also that it is close enough to a sufficient amount of gateways.
		numGatewaysCloseEnough := 0
		gatewayCloseToSensor := make([]int, len(gateways))

		for j := 0; j < len(gateways); j++ {

			d := calculateDistance(p, gateways[j])

			// Sensor is too close to a gateway.
			if d < minDistToGw {
				attempts++
				goto Top
			}

			// Sensor is in range of gateway.
			if d < maxDistToGw {
				numGatewaysCloseEnough++
				gatewayCloseToSensor[j] = 1
			}
		}

		// See if the sensor is close enough to a sufficient amount of gateways.
		if numGatewaysCloseEnough < minGwPerSensor {
			attempts++
			goto Top
		}

		// See if there are too many sensors close to any given gateway.
		for k := range gatewayCloseToSensor {
			if gatewayCloseToSensor[k]+sensorsPerGateway[k] > maxSensorPerGw {
				attempts++
				goto Top
			}
		}

		// If we got here we successfully made it!
		positions = append(positions, p)
		for k := range gatewayCloseToSensor {
			sensorsPerGateway[k] += gatewayCloseToSensor[k]
		}
		i++

		attempts++
	}

	// Check if any gateways are without sufficient amount of sensors
	for k := range sensorsPerGateway {
		if sensorsPerGateway[k] < minSensorsPerGw {
			return nil, fmt.Errorf("Gateway missing sensors. Try again.")
		}
	}

	return positions, nil
}

func getUserInput() settings {
	//reader := bufio.NewReader(os.Stdin)

	var inputs = map[int]string{
		0:  "Grid width",
		1:  "Grid height",
		2:  "Number of gateways",
		3:  "Number of sensors",
		4:  "Minimum distance gateway-gateway",
		5:  "Minimum distance sensor-sensor",
		6:  "Minimum distance sensor-gateway",
		7:  "Maximum distance sensor-gateway",
		8:  "Maximum number of sensors per gateway",
		9:  "Minimum number of gateways per sensor",
		10: "Minimum number of sensors per gateway",
		11: "Number of grids to create",
		12: "Filename",
	}

	s := settings{}

	for i := 0; i < len(inputs); i++ {

		fmt.Printf("%s: ", inputs[i])

		switch i {
		case 0:
			s.width = readInt()

		case 1:
			s.height = readInt()

		case 2:
			s.numGateways = readInt()

		case 3:
			s.numSensors = readInt()

		case 4:
			s.minDistanceGwGw = readInt()

		case 5:
			s.minDistanceSensorSensor = readInt()

		case 6:
			s.minDistanceSensorGw = readInt()

		case 7:
			s.maxDistanceSensorGw = readInt()

		case 8:
			s.maxSensorPerGw = readInt()

		case 9:
			s.minGwPerSensor = readInt()

		case 10:
			s.minSensorPerGw = readInt()

		case 11:
			s.numGrids = readInt()

		case 12:
			s.filename = readString()
		}
	}
	return s
}

func readInt() int {
	var num int
	_, err := fmt.Scanf("%d\n", &num)
	if err != nil {
		panic(err)
	}
	return num
}

func readString() string {
	var s string
	_, err := fmt.Scanf("%s\n", &s)
	if err != nil {
		panic(err)
	}
	return s
}

func calculateDistance(a, b position) float64 {
	return math.Sqrt(math.Pow(float64(a.X-b.X), 2) + math.Pow(float64(a.Y-b.Y), 2))
}
