package benchmark

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"path/filepath"

	"github.com/orjansv/master-code/simulator"
	"github.com/orjansv/master-code/utils"
)

// Benchmark is a struct used to hold stats about a single benchmark.
type Benchmark struct {
	Method                 string
	TotalRetries           int
	TotalEnergyConsumption float64
	TotalBoost             int
	Retries                []int
	Energy                 []float64
	Boosts                 []int
}

type BenchmarkInfo struct {
	Name                 string
	UsesMultiCoordinates bool
	NumCoordinates       int
	SensorCoordinates    [][]utils.Coordinates
	GatewayCoordinates   [][]utils.Coordinates
	Title                string
}

func GetBenchmarkStats(filename string) BenchmarkInfo {

	b := BenchmarkInfo{}
	var extension = filepath.Ext(filename)

	if extension != ".json" {
		err := fmt.Sprintf("Configuration file must be a json file. Have a %s file.\n", extension)
		panic(err)
	}

	b.Name = filename[0 : len(filename)-len(extension)]

	file, err := ioutil.ReadFile(filename)
	if err != nil {
		panic(err)
	}

	var conf simulator.ConfigJSON
	err = json.Unmarshal(file, &conf)
	if err != nil {
		error := fmt.Sprintf("Failed to load config file: %s", err)
		panic(error)
	}

	// Check for multiple coordinates
	numMultiCoordSensor := len(conf.Sensors.MultiCoordinates)
	numMultiCoordGw := len(conf.Gateways.MultiCoordinates)

	if numMultiCoordGw != numMultiCoordSensor {
		if numMultiCoordGw > 0 && numMultiCoordSensor > 0 {
			fmt.Println("WARNING: uneven amount of multi-coordinates for sensors and gateways. Will only use one.")

			b.SensorCoordinates = append(b.SensorCoordinates, conf.Sensors.MultiCoordinates[0])
			b.GatewayCoordinates = append(b.GatewayCoordinates, conf.Gateways.MultiCoordinates[0])
			b.UsesMultiCoordinates = true
			b.NumCoordinates = 1
		} else {
			b.UsesMultiCoordinates = false
		}
	} else if numMultiCoordGw > 0 {
		b.SensorCoordinates = conf.Sensors.MultiCoordinates
		b.GatewayCoordinates = conf.Gateways.MultiCoordinates
		b.UsesMultiCoordinates = true
		b.NumCoordinates = numMultiCoordGw
	}
	if len(conf.Simulation.Title) > 0 {
		b.Title = conf.Simulation.Title
	} else {
		b.Title = b.Name
	}

	return b
}
