package cloud

import (
	"fmt"
	"math"
	"time"

	"github.com/orjansv/master-code/simulator/common"
	"github.com/orjansv/master-code/simulator/simtime"
	"github.com/orjansv/master-code/utils"
)

// Cloud is the type that simulates the cloud.
type Cloud struct {
	channels           Channels
	messages           map[MessageKey]*Event
	retries            map[ConnectionKey]*utils.MovingWindowInt
	selectionMethod    int
	startTime          time.Time
	endTime            simtime.SimulatedTime
	responsibleGateway map[int]*ResponsibleGateway
	upLink             UpLink
	upLinkBoost        UpLink
	sensor             SensorInfo
	connections        map[ConnectionKey]*ConnectionStatus
	windowSize         int
	downlinks          map[ConnectionKey]*DownLink
	downlinksBoost     map[ConnectionKey]*DownLink
	config             Config
	reason             string
	changeGwChan       chan int
	verboseLevel       int
	bitCost            float64
	boostCost          float64
	sleepTime          time.Duration
}

// NewCloud returns a new instance of the cloud type.
func NewCloud(channels Channels, decisionType int, s SensorInfo, duration simtime.SimulatedTime, windowSize int, conf Config, verboseLevel int, bitCost, boostCost float64, boostEnabled bool) Cloud {

	messages := make(map[MessageKey]*Event)

	retries := make(map[ConnectionKey]*utils.MovingWindowInt)

	responsibleGateway := make(map[int]*ResponsibleGateway)

	// Convert txPower from dBm to mW (TODO: extract to function?)
	txPowermW := make([]float64, len(s.TransmissionPowers))
	for i, val := range s.TransmissionPowers {
		txPowermW[i] = math.Pow(10, val/10)
	}
	s.TransmissionPowers = txPowermW

	upLink := UpLink{}
	upLink.MessagesReceivedTotal = make(map[int]*utils.MovingWindowFloat)
	upLink.MessagesReceivedConnection = make(map[ConnectionKey]*utils.MovingWindowFloat)
	upLink.Available = make(map[ConnectionKey]bool)

	upLinkBoost := UpLink{}
	upLinkBoost.MessagesReceivedTotal = make(map[int]*utils.MovingWindowFloat)
	upLinkBoost.MessagesReceivedConnection = make(map[ConnectionKey]*utils.MovingWindowFloat)
	upLinkBoost.Available = make(map[ConnectionKey]bool)

	connections := make(map[ConnectionKey]*ConnectionStatus)

	downlinks := make(map[ConnectionKey]*DownLink)
	downlinksBoost := make(map[ConnectionKey]*DownLink)

	changeGwChan := make(chan int, 100) //TODO: dynamic sizing (one for each sensor?)

	sleeptime := time.Duration(s.MaxAttempts+1) * s.WaitingTime
	if boostEnabled {
		sleeptime = time.Duration(s.MaxAttempts*2+1) * simtime.Minute
	}

	c := Cloud{
		channels,
		messages,
		retries,
		decisionType,
		time.Now(),
		duration,
		responsibleGateway,
		upLink,
		upLinkBoost,
		s,
		connections,
		windowSize,
		downlinks,
		downlinksBoost,
		conf,
		selectionReasons[decisionType],
		changeGwChan,
		verboseLevel,
		bitCost,
		boostCost,
		sleeptime,
	}

	return c
}

// Run starts the simulation of the cloud.
func (c *Cloud) Run() {

	c.startTime = time.Now()

	checkDuration := false
	// Check duration if endtime is set (i.e. it is greater than 0).
	noTime := simtime.SimulatedTime{Days: 0, Hours: 0, Minutes: 0, Seconds: 0}
	if c.endTime.SubInMinutes(noTime) > 0 {
		checkDuration = true
	}

	// Listen for messages and handle them properly.
	var day int64
	for {
		select {
		case msg := <-c.channels.InChannel:
			// TODO: currently using timestamp in cloud. Should use from gateway!
			timePassed := simtime.GetSimulationTime(c.startTime)
			timestamp := timePassed.ToString()

			if timePassed.Days != day {
				if c.verboseLevel > 0 {
					fmt.Printf("Day %d\n", timePassed.Days)
				}
				day = timePassed.Days
			}

			// This order is important for retry detection!
			c.HandleMessage(msg)
			c.InsertMessage(msg, timestamp)
			c.PrintMessage(msg, timestamp)

			// Stop if we reached end time.
			if checkDuration == true {
				if c.endTime.SubInHours(timePassed) <= 0 {
					fmt.Printf("Simulation execution time reached.\n\n")
					c.stop()
					return
				}
			}

		// Triggered by a timer
		case changeID := <-c.changeGwChan:
			c.ChangeOptimalGateway(changeID)

		case <-c.channels.StopChan:
			// User told us to stop
			c.stop()

		}
	}
}

func (c *Cloud) stop() {
	allMessages := c.prepareMessages()

	// Send messages to simulator
	c.channels.MessagesChan <- allMessages
}

// ChangeOptimalGateway changes the optimal gateway for sending replies via.
func (c *Cloud) ChangeOptimalGateway(sensorID int) {

	newKey := ConnectionKey{sensorID, c.responsibleGateway[sensorID].NextResponsible}
	oldKey := ConnectionKey{sensorID, c.responsibleGateway[sensorID].ID}

	// Only change gateway if it's a new one.
	if c.responsibleGateway[newKey.SensorID].NextResponsible != oldKey.GatewayID {
		// Flush replies from previous gateway
		c.channels.OutChans[oldKey.GatewayID] <- common.CommandMessage{Command: common.CMDFlush, SensorID: oldKey.SensorID}

		// Set new responsible gateway
		c.responsibleGateway[newKey.SensorID].ID = newKey.GatewayID
		c.responsibleGateway[newKey.SensorID].RepliesAvailable = 0
		c.responsibleGateway[newKey.SensorID].Replied = false

		printGatewayChanged(newKey.SensorID, newKey.GatewayID, c.reason, c.verboseLevel)
	}
	c.SendReplies(sensorID, c.sensor.MaxAttempts*2-c.responsibleGateway[sensorID].RepliesAvailable) // TODO: amount of messages.
}

// SendReplies sends pre-made replies to a specified gateway for a specific sensor.
func (c *Cloud) SendReplies(sensorID, numReplies int) {

	reply := common.CommandMessage{Command: common.CMDReply, SensorID: sensorID, NumReplies: numReplies}
	select {
	case c.channels.OutChans[c.responsibleGateway[sensorID].ID] <- reply:
		// Sent reply.
	default:
		if c.verboseLevel > 0 {
			fmt.Println("WRN Could not send message with replies to gateway!")
		}
	}
}

// ChangeOptimalGatewayNew changes the optimal gateway for sending replies via.
func (c *Cloud) startChangeGwTimer(sensorID int) {
	time.Sleep(c.sleepTime)
	c.changeGwChan <- sensorID
}
