package cloud

import (
	"math"

	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/utils"
)

func newConnectionStatus(windowSize int) *ConnectionStatus {
	return &ConnectionStatus{
		SignalStrength:            utils.NewMAW(windowSize),
		ExpectedEnergyConsumption: utils.NewMAW(windowSize),
	}
}

// Reliability calculates the rate of how many messages are received per connection out of the total amount of messages received.
func (u UpLink) Reliability(key ConnectionKey) float64 {
	// Out of all messages, how many were received on this specific connection?
	if u.Available[key] {
		if u.MessagesReceivedTotal[key.SensorID].Total() == 0 {
			// No messages have been received on the connection during the window size.
			// But it has previously been seen, so we might give it a chance by assuming it is now reliable.
			return 1
		}
		return u.MessagesReceivedConnection[key].Total() / u.MessagesReceivedTotal[key.SensorID].Total()
	}

	// Connection has never been seen.
	return 0

}

// Reliability calculates the probability that downlink messages are successfully received by the sensor.
func (d DownLink) Reliability() float64 {
	if d.retriesCaused.Total() <= 0 || d.repliesSent.Total() == 0 {
		// Downlink has not been used, so we assume it is reliable until otherwise proven.
		return 1
	}
	return 1.0 - d.retriesCaused.Total()/d.repliesSent.Total()
}

func (c *Cloud) updateAverageSignalStrength(k ConnectionKey, msg sensor.Message) {
	// Calculate average signal strength for connection
	c.connections[k].SignalStrength.Insert(msg.SignalStrength)
}

func (c *Cloud) checkInitialization(k ConnectionKey, msg sensor.Message) {

	// Check that boost maps and structs are initialized.
	if msg.Boost {
		if c.upLinkBoost.Available[k] != true {
			// New connection, so we must initialize.
			c.upLinkBoost.Available[k] = true
			c.upLinkBoost.MessagesReceivedConnection[k] = utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0)

			if _, ok := c.responsibleGateway[msg.SensorID]; !ok {
				// First time this sensor is seen! We must start from -1 to cope for initial retry and sent reply, as this gateway will now be responsible.
				c.downlinksBoost[k] = &DownLink{
					retriesCaused: utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, -1),
					repliesSent:   utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, -1),
				}
			} else {
				// Sensor has been seen before. Initialize this new downlink to 0 sent and 0 retries (because it has never been used as responsible before).
				c.downlinksBoost[k] = &DownLink{
					retriesCaused: utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0),
					repliesSent:   utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0),
				}

			}
		}
	} else {
		if c.upLink.Available[k] != true {
			// New connection, so we must initialize.
			c.upLink.Available[k] = true
			c.upLink.MessagesReceivedConnection[k] = utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0)

			if _, ok := c.responsibleGateway[msg.SensorID]; !ok {
				// First time this sensor is seen! We must start from -1 to cope for initial retry and sent reply, as this gateway will now be responsible.
				c.downlinks[k] = &DownLink{
					retriesCaused: utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, -1),
					repliesSent:   utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, -1),
				}

			} else {
				// Sensor has been seen before. Initialize new downlink to 0 sent and 0 retries (because it has never been used as responsible before).
				c.downlinks[k] = &DownLink{
					retriesCaused: utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0),
					repliesSent:   utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0),
				}
			}
		}
	}

	// Initialize shared maps and structs.
	if _, ok := c.connections[k]; !ok {
		// New connection!
		c.connections[k] = newConnectionStatus(c.windowSize)

		if _, ok := c.responsibleGateway[msg.SensorID]; !ok {
			// First time this sensor is seen! We must start from -1 to cope for initial retry and sent reply.
			c.retries[k] = utils.NewMovingWindowInt(msg.MessageID, c.windowSize, -1)

			// Set this gateway as responsible, since it is the first one available.
			c.responsibleGateway[msg.SensorID] = &ResponsibleGateway{msg.GatewayID, 0, false, false, msg.GatewayID}
			printGatewayChanged(msg.SensorID, msg.GatewayID, "no previous optimal gateway.", c.verboseLevel) // TODO: move/delete

			// Uplinks
			c.upLink.MessagesReceivedTotal[msg.SensorID] = utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0)
			c.upLinkBoost.MessagesReceivedTotal[msg.SensorID] = utils.NewMovingWindowFloat(msg.MessageID, c.windowSize, 0)

		} else {
			// Sensor has been seen before. Initialize retries to 0.
			c.retries[k] = utils.NewMovingWindowInt(msg.MessageID, c.windowSize, 0)
		}
	}

}

func (c *Cloud) updateExpectedEnergyConsumption(msg sensor.Message, key ConnectionKey) {
	// Calculate expected energy consumption based on reliability.
	expectedEnergyConsumption := c.calculateExpectedEnergyConsumption(msg)

	// Add expected energy consumption to the average window
	c.connections[key].ExpectedEnergyConsumption.Insert(expectedEnergyConsumption)
}

func (c *Cloud) calculateExpectedEnergyConsumption(msg sensor.Message) float64 {
	connectionKey := ConnectionKey{msg.SensorID, msg.GatewayID}

	expectedEnergyConsumption := 0.0
	Pr := 1.0 // Assume that normal connection is unavailable (initially).
	if c.upLink.Available[connectionKey] {

		// Energy consumption for first transmission.
		estimatedEnergyConsumption := c.sensor.TransmissionPowers[msg.TransmissionLevel] * 8 * c.bitCost * msg.PacketSize / c.sensor.DataRate
		Pr = (1.0 - c.upLink.Reliability(connectionKey)) + (1.0-c.downlinks[connectionKey].Reliability())*c.upLink.Reliability(connectionKey)

		// Energy consumption for retries
		expectedEnergyConsumption += estimatedEnergyConsumption
		for i := 1; i < c.sensor.MaxAttempts; i++ {
			expectedEnergyConsumption += estimatedEnergyConsumption * math.Pow(Pr, float64(i))
		}
	}

	// Only calculate for boost if it is available on this connection.
	if c.upLinkBoost.Available[connectionKey] {
		// Energy consumption for one boost transmission.
		estimatedEnergyConsumptionBoost := c.sensor.TransmissionPowers[msg.TransmissionLevel] * 8 * c.bitCost * msg.PacketSize * c.boostCost / c.sensor.DataRate
		PrBoost := (1.0 - c.upLinkBoost.Reliability(connectionKey)) + (1.0-c.downlinksBoost[connectionKey].Reliability())*c.upLinkBoost.Reliability(connectionKey)

		// Initial transmission in boost
		expectedEnergyConsumption += Pr * estimatedEnergyConsumptionBoost

		// Energy consumption for retries in boost
		for i := 1; i < c.sensor.MaxAttempts; i++ {
			expectedEnergyConsumption += estimatedEnergyConsumptionBoost * math.Pow(PrBoost, float64(i)) * math.Pow(Pr, float64(c.sensor.MaxAttempts+1))
		}
	}

	return expectedEnergyConsumption
}

func (c *Cloud) monitorUplink(msg sensor.Message) {
	connectionKey := ConnectionKey{msg.SensorID, msg.GatewayID}

	if msg.Boost {
		// First, check if we should count it as a completely new message (total).
		if c.MessageIsNew(msg) {
			c.upLinkBoost.MessagesReceivedTotal[msg.SensorID].Insert(msg.MessageID)
		}

		c.upLinkBoost.MessagesReceivedConnection[connectionKey].Insert(msg.MessageID)

		// If the uplink for this connection has previously been available in normal mode,
		// we'll move its window so it might get a new chance at being responsible if its
		// reliability changes.
		if c.upLink.Available[connectionKey] {
			c.upLink.MessagesReceivedConnection[connectionKey].Move(msg.MessageID)
			c.upLink.MessagesReceivedTotal[msg.SensorID].Move(msg.MessageID)
		}

	} else {
		if c.MessageIsNew(msg) {
			c.upLink.MessagesReceivedTotal[msg.SensorID].Insert(msg.MessageID)
		}

		c.upLink.MessagesReceivedConnection[connectionKey].Insert(msg.MessageID)

		// If the uplink for this connection has previously been available in boost mode,
		// we'll move its window so it might get a new chance at being responsible if its
		// reliability changes.
		if c.upLinkBoost.Available[connectionKey] {
			c.upLinkBoost.MessagesReceivedConnection[connectionKey].Move(msg.MessageID)
			c.upLinkBoost.MessagesReceivedTotal[msg.SensorID].Move(msg.MessageID)
		}
	}
}

func (c *Cloud) monitorDownlink(msg sensor.Message, newKey, optimalKey ConnectionKey) {

	msgKey := MessageKey{msg.SensorID, msg.MessageID}
	if _, ok := c.messages[msgKey]; !ok {
		// New message, remove replied stamp
		c.responsibleGateway[msg.SensorID].Replied = false
		c.responsibleGateway[msg.SensorID].RepliedBoost = false
	} else if c.MessageIsRetry(msg) {

		// If the gateway did reply to the last attempt, blame its downlink for the retry.
		if c.responsibleGateway[msg.SensorID].Replied == true {
			// Shouldn't really be necessary.
			if c.upLink.Available[optimalKey] {
				c.downlinks[optimalKey].retriesCaused.Insert(msg.MessageID)
			}
			c.responsibleGateway[msg.SensorID].Replied = false
		} else if c.responsibleGateway[msg.SensorID].RepliedBoost == true {
			// Shouldn't really be necessary.
			if c.upLinkBoost.Available[optimalKey] {
				c.downlinksBoost[optimalKey].retriesCaused.Insert(msg.MessageID)
			}
			c.responsibleGateway[msg.SensorID].RepliedBoost = false
		}
	}

	// If the optimal gateway forwarded a message to the cloud, it also replied to the sensor.
	if newKey.GatewayID == c.responsibleGateway[msg.SensorID].ID {

		if msg.Boost {
			// Shouldn't really be necessary.
			if c.upLinkBoost.Available[newKey] {
				c.downlinksBoost[newKey].repliesSent.Insert(msg.MessageID)
			}
			c.responsibleGateway[msg.SensorID].RepliedBoost = true
		} else {
			// Shouldn't really be necessary.
			if c.upLink.Available[newKey] {
				c.downlinks[newKey].repliesSent.Insert(msg.MessageID)
			}
			c.responsibleGateway[msg.SensorID].Replied = true
		}
	}

}

func (c *Cloud) calculateWeightedReliability(key ConnectionKey, boostEnabled bool) float64 {
	if boostEnabled {
		if c.upLinkBoost.Available[key] {
			if c.upLink.Available[key] {
				// Both connections have been seen and should be used in the calculation.
				return c.config.WeightUplink*(c.upLink.Reliability(key)+c.upLinkBoost.Reliability(key))/2 +
					c.config.WeightDownlink*(c.downlinks[key].Reliability()+c.downlinksBoost[key].Reliability())/2
			}

			// Only boost has been seen, and should be used.
			return c.config.WeightUplink*c.upLinkBoost.Reliability(key) + c.config.WeightDownlink*c.downlinksBoost[key].Reliability()
		}

		// Only normal connection has been seen and should be used.
		return c.config.WeightUplink*c.upLink.Reliability(key) + c.config.WeightDownlink*c.downlinks[key].Reliability()

	}

	// Only use normal connections.
	return c.config.WeightUplink*c.upLink.Reliability(key) + c.config.WeightDownlink*c.downlinks[key].Reliability()
}
