package simulator

import (
	"fmt"
	"math"

	"github.com/orjansv/master-code/simulator/cloud"
	"github.com/orjansv/master-code/simulator/common"
	"github.com/orjansv/master-code/simulator/connection"
	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/simulator/watchdog"
	"github.com/orjansv/master-code/utils"
)

/*
	Each row in the matrix represents the distances from one sensor to all the gateways.
	|    | G1 | G2   |
	| S1 |  1 |  2.5 |
*/
func (s *Simulator) getDistanceMatrix() [][]float64 {

	distanceMatrix := make([][]float64, s.config.numSensors)

	for i := 0; i < s.config.numSensors; i++ {
		distanceMatrix[i] = make([]float64, s.config.numGateways)
		for j := 0; j < s.config.numGateways; j++ {
			distanceMatrix[i][j] = utils.CalculateDistance(s.config.sensors.positions[i], s.config.gateways.positions[j])
		}
	}

	return distanceMatrix
}

/*
 Connections are shown as as a boolean matrix:
 |	  | G1 | G2 |
 | S1 |  1 |  0 |
*/
func (s *Simulator) getConnectionMatrix(distanceMatrix [][]float64, upper, lower float64) ([][]int, int) {

	numConnections := 0
	connectionMatrix := make([][]int, s.config.numSensors)
	sensorsWithNoConnection := 0

	for i := 0; i < s.config.numSensors; i++ {
		connectionMatrix[i] = make([]int, s.config.numGateways)
		numConnForSensor := 0
		for j := 0; j < s.config.numGateways; j++ {
			if distanceMatrix[i][j] <= upper && distanceMatrix[i][j] > lower {
				connectionMatrix[i][j] = 1
				numConnections++
				numConnForSensor++
			}
		}
		if numConnForSensor == 0 {
			sensorsWithNoConnection++
		}
	}

	if sensorsWithNoConnection == 1 {
		fmt.Printf("WARNING: %d sensor has no connections at all (except boost)!\n", sensorsWithNoConnection)
	} else if sensorsWithNoConnection > 1 {
		fmt.Printf("WARNING: %d sensors have no connections at all (except boost)!\n", sensorsWithNoConnection)
	}

	return connectionMatrix, numConnections

}

func (s *Simulator) makeChannels() {
	// Create channel used for stopping the simulation.
	s.stopChannel = make(chan string)

	sensor2GatewayChans := make([]chan sensor.Message, s.config.numGateways)
	for i := range sensor2GatewayChans {
		sensor2GatewayChans[i] = make(chan sensor.Message)
	}

	gateway2SensorChans := make([]chan sensor.Message, s.config.numSensors)
	for i := range gateway2SensorChans {
		gateway2SensorChans[i] = make(chan sensor.Message)
	}

	sensorStopChans := make([]chan int, s.config.numSensors)
	for i := range sensorStopChans {
		sensorStopChans[i] = make(chan int)
	}

	gatewayStopChans := make([]chan int, s.config.numGateways)
	for i := range gatewayStopChans {
		gatewayStopChans[i] = make(chan int)
	}

	s.channels.cloud2GatewayChans = make([]chan common.CommandMessage, s.config.numGateways)
	s.channels.gateway2CloudChan = make(chan sensor.Message, 1000) // TODO: num sensors? * 10?
	s.channels.gateway2SensorChans = gateway2SensorChans
	s.channels.sensor2GatewayChans = sensor2GatewayChans
	s.channels.watchdogMessageChan = make(chan watchdog.WatchdogMessage, 1000) // for sensor to tell watchdog it sent a retry
	s.channels.watchDogStopChan = make(chan string)
	s.channels.messagesChan = make(chan cloud.JSONTop)
	s.channels.energyStatsChan = make(chan watchdog.EnergyStats)
	s.channels.sensorStopChans = sensorStopChans
	s.channels.gatewayStopChans = gatewayStopChans

	// Set up channels from the cloud to all gateways
	for i := range s.channels.cloud2GatewayChans {
		s.channels.cloud2GatewayChans[i] = make(chan common.CommandMessage, 1000)
	}
}

/* checkConnectionDistribution
** The connections between sensors and gateways can be of several types.
** The distribution of the types is determined in the configuration file.
** If it is not set properly, it defaults to the standard connection.
 */
func (s *Simulator) checkConnectionDistribution() {

	// Ensure that total distribution of connections sums up to 1.
	// If not, add the missing to the default connection
	distributionSum := 0.0
	for i := 0; i < numConnectionTypes; i++ {
		distributionSum += s.connectionDistribution[i]
	}
	s.connectionDistribution[connection.TypeDefault] += (1 - distributionSum)

	distributionSum = 0.0
	for i := 0; i < numConnectionTypes; i++ {
		distributionSum += s.boostConnectionDistribution[i]
	}
	s.boostConnectionDistribution[connection.TypeDefault] += (1 - distributionSum)
}

/* getConnectionTypes
** returns a slice with one type for all connections, shuffled randomly.
 */
func (s *Simulator) getConnectionTypes(numConnections int, distribution []float64) []int {

	// Probabilities for connection types as gotten from file
	connectionTypes := make([]int, numConnections)

	numTypes := make([]int, numConnectionTypes)
	sum := 0
	for i := 0; i < numConnectionTypes; i++ {
		numTypes[i] = int(distribution[i] * float64(numConnections))
		sum += numTypes[i]
	}

	// If we did not fill enough, iterate until we have given types for all connections.
	numToFill := numConnections - sum
	if numToFill != 0 {
		i := 0
		filled := 0
		for {
			if numTypes[i] != 0 {
				numTypes[i]++
				filled++
			}

			i = (i + 1) % numConnectionTypes

			if filled == numToFill {
				break
			}
		}
	}

	index := 0
	for i := 0; i < numConnectionTypes; i++ {
		for j := 0; j < numTypes[i]; j++ {
			connectionTypes[index] = i
			index++
		}
	}

	connectionTypes = utils.Shuffle(connectionTypes, s.config.seed)
	return connectionTypes
}

// Returns the distance threshold based on Free-space path loss.
func (s *Simulator) calculateThreshold() float64 {

	// Max distance from sensor's perspective.
	sensorThreshold := math.Pow(10, ((s.config.gateways.transmissionPower - 20*math.Log10(s.config.frequency) + 147.55 - s.config.sensors.sensitivity) / 20.0))

	// Max distance from gateway's perspective.
	gatewayThreshold := math.Pow(10, ((s.config.sensors.transmissionPower[len(s.config.sensors.transmissionPower)-1] - 20*math.Log10(s.config.frequency) + 147.55 - s.config.gateways.sensitivity) / 20))

	if gatewayThreshold < sensorThreshold {
		return gatewayThreshold
	}
	return sensorThreshold
}
