package simulator

import (
	"bufio"
	"fmt"
	"os"
	"time"

	"github.com/orjansv/master-code/simulator/cloud"
	"github.com/orjansv/master-code/simulator/connection"
	"github.com/orjansv/master-code/simulator/gateway"
	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/simulator/watchdog"
	"github.com/orjansv/master-code/utils"
)

// Simulator is the type used to tie together the whole simulation.
type Simulator struct {
	stopChannel                 chan string
	VerboseLevel                int
	connectionDistribution      []float64
	boostConnectionDistribution []float64
	config                      simulatorConfig
	channels                    simulatorChannels
	cloud                       *cloud.Cloud
	watchdog                    *watchdog.Watchdog
	gateways                    []*gateway.Gateway
	sensors                     []*sensor.Sensor
	gridJSON                    GridJSON
}

// Run starts the simulation of the whole system.
func (s *Simulator) Run() (SimulationResults, error) {

	res := s.simulate()

	return res, nil
}

// Setup initializes the simulation.
func (s *Simulator) Setup(seed int) {

	// General setup.
	distanceMatrix := s.getDistanceMatrix()
	connectionMatrix, numConnections := s.getConnectionMatrix(distanceMatrix, s.config.normalRange, 0)
	s.checkConnectionDistribution()
	connectionTypes := s.getConnectionTypes(numConnections, s.connectionDistribution)

	// Update seed after initializing connections, to make sure the connections are always the same.
	s.config.seed += int64(seed)

	// Set up boost connections and add them to JSON.
	var connectionMatrixBoost [][]int
	var numBoostConnections int
	if s.config.boostEnabled {
		connectionMatrixBoost, numBoostConnections = s.getConnectionMatrix(distanceMatrix, s.config.boostRange, 0)
	}
	boostConnectionTypes := s.getConnectionTypes(numBoostConnections, s.boostConnectionDistribution)

	s.addConnectionsToJSON(connectionTypes, connectionMatrix, connectionMatrixBoost)

	s.makeChannels()

	// Setup components.
	s.cloud = s.setupCloud()
	s.watchdog = s.setupWatchDog()
	s.gateways = s.setupGateways(connectionMatrix, connectionMatrixBoost)
	s.sensors = s.setupSensors(connectionMatrix, connectionMatrixBoost, distanceMatrix, connectionTypes, boostConnectionTypes)
}

func (s *Simulator) simulate() SimulationResults {

	fmt.Println("Selection method:", cloud.SelectionMethodsString[s.config.selectionMethod])

	// Start all components.
	go s.cloud.Run()
	go s.watchdog.Run()

	for _, g := range s.gateways {
		go g.Run()
	}
	// Start sensors with delay in between.
	for _, sensor := range s.sensors {
		time.Sleep(s.config.sensors.startInterval)
		go sensor.Run()
	}

	// Stop code on key press, or wait for simulation to finish.
	inputChan := make(chan string)
	go getUserInput(inputChan)

	var messages cloud.JSONTop
	select {
	case messages = <-s.channels.messagesChan:
		// The cloud timed out and sent us its messages.
		// Now let's stop everything else in the simulator.

	case <-inputChan:
		// User made simulation stop. Now stop all components properly.
		s.stopChannel <- "stop"

		// Wait for messages from cloud.
		messages = <-s.channels.messagesChan
	}

	// Stop watchdog.
	s.channels.watchDogStopChan <- "stop"

	// Wait for energy info from watchdog.
	energyStats := <-s.channels.energyStatsChan

	// Stop all sensors.
	for i := range s.channels.sensorStopChans {
		s.channels.sensorStopChans[i] <- 1
	}

	// Wait for confirmation that they have stopped.
	for i := range s.channels.sensorStopChans {
		<-s.channels.sensorStopChans[i]
	}

	// Stop all gateways.
	for i := range s.channels.gatewayStopChans {
		s.channels.gatewayStopChans[i] <- 1
	}

	// Wait for confirmation that they have stopped.
	for i := range s.channels.gatewayStopChans {
		<-s.channels.gatewayStopChans[i]
	}

	return SimulationResults{
		Messages:    messages,
		Performance: energyStats,
		Connections: s.gridJSON,
	}
}

func getUserInput(input chan string) {
	reader := bufio.NewReader(os.Stdin)
	_, _ = reader.ReadString('\n')
	input <- "stop"
}

// SetPositions overwrites the positions read from the configuration file.
func (s *Simulator) SetPositions(sensors, gateways []utils.Coordinates) {
	s.config.sensors.positions = sensors
	s.config.numSensors = len(sensors)
	s.config.gateways.positions = gateways
	s.config.numGateways = len(gateways)

	// Update start interval
	s.config.sensors.startInterval = s.config.sensors.sendingInterval / time.Duration(s.config.numSensors)

	s.gridJSON.Sensors = sensors
	s.gridJSON.Gateways = gateways
}

// SetSelectionMethod overrides the gateway selection type read from the configuration file.
func (s *Simulator) SetSelectionMethod(method int) {

	s.config.selectionMethod = method
	if method == cloud.SelectionMethodLowestEnergyConsumption && s.config.sensors.sensorType == sensor.TypeDynamic {
		s.config.sensors.sendTPC = true
	}

}

func (s *Simulator) addConnectionsToJSON(connectionTypes []int, connectionMatrix, connectionMatrixBoost [][]int) {

	connTypeCounter := 0
	for i := 0; i < s.config.numSensors; i++ {
		for j := 0; j < s.config.numGateways; j++ {
			if connectionMatrix[i][j] == 1 {
				switch connectionTypes[connTypeCounter] {
				case connection.TypeDefault:
					s.gridJSON.Connections.Standard = append(s.gridJSON.Connections.Standard,
						LinesJSON{
							X: [2]float64{s.config.sensors.positions[i].X, s.config.gateways.positions[j].X},
							Y: [2]float64{s.config.sensors.positions[i].Y, s.config.gateways.positions[j].Y}})

				case connection.TypeDynamic:
					s.gridJSON.Connections.Dynamic = append(s.gridJSON.Connections.Dynamic,
						LinesJSON{
							X: [2]float64{s.config.sensors.positions[i].X, s.config.gateways.positions[j].X},
							Y: [2]float64{s.config.sensors.positions[i].Y, s.config.gateways.positions[j].Y}})

				case connection.TypeBinary:
					s.gridJSON.Connections.Binary = append(s.gridJSON.Connections.Binary,
						LinesJSON{
							X: [2]float64{s.config.sensors.positions[i].X, s.config.gateways.positions[j].X},
							Y: [2]float64{s.config.sensors.positions[i].Y, s.config.gateways.positions[j].Y}})
				}

				connTypeCounter++
			}

			if s.config.boostEnabled {
				if connectionMatrixBoost[i][j] == 1 {
					s.gridJSON.Connections.Boost = append(s.gridJSON.Connections.Boost,
						LinesJSON{
							X: [2]float64{s.config.sensors.positions[i].X, s.config.gateways.positions[j].X},
							Y: [2]float64{s.config.sensors.positions[i].Y, s.config.gateways.positions[j].Y},
						})
				}
			}
		}
	}
}
