package watchdog

import (
	"fmt"
	"math"
)

// Watchdog is the type used for monitoring retries and energy usage.
type Watchdog struct {
	messageChan         chan WatchdogMessage
	stopChan            chan string
	energyChan          chan EnergyStats
	transmissionPowermW []float64
	dataRate            float64
	bitCost             float64
	boostCost           float64
}

// WatchdogMessage is the message sent from sensors to the watchdog.
type WatchdogMessage struct {
	SensorID               int
	NumBytes               int
	TransmissionPowerLevel int
	Retry                  bool
	Boost                  bool
}

// EnergyStats is used to hold information about a simulation.
type EnergyStats struct {
	TotalRetries           int
	TotalEnergyConsumption float64
	TotalBoostMessages     int
	Sensors                []SensorEnergyStats
}

// SensorEnergyStats is used to hold energy stats for an individual sensor.
type SensorEnergyStats struct {
	SensorID          int     `json:"id"`
	EnergyConsumption float64 `json:"energy_consumption"`
	Retries           int     `json:"retries"`
}

// NewWatchDog returns a new instance of the watchdog type.
func NewWatchDog(messageChan chan WatchdogMessage, stopChan chan string, energyChan chan EnergyStats, sensorTransmissionPowerLevels []float64, dataRate, bitCost, boostCost float64) Watchdog {

	txPowermW := make([]float64, len(sensorTransmissionPowerLevels))
	for i, val := range sensorTransmissionPowerLevels {
		txPowermW[i] = math.Pow(10, val/10)
	}

	return Watchdog{
		messageChan,
		stopChan,
		energyChan,
		txPowermW,
		dataRate,
		bitCost,
		boostCost,
	}
}

// Run starts the watchdog.
func (w *Watchdog) Run() {

	numRetries := make(map[int]int)
	energyConsumption := make(map[int]float64)
	//messagesSent := 0
	//avgTransmLvl := 0.0
	//retrycost := 0.0
	boostMessages := 0

Loop:
	for {

		select {
		case msg := <-w.messageChan:
			if msg.Retry == true {
				numRetries[msg.SensorID]++
				//retrycost += w.calculateEnergy(msg.NumBytes, msg.TransmissionPowerLevel, 1)
			}

			//messagesSent++
			//avgTransmLvl = (avgTransmLvl + float64(msg.TransmissionPowerLevel)) / 2.0

			if msg.Boost {
				boostMessages++
				energyConsumption[msg.SensorID] += w.calculateEnergy(msg.NumBytes, msg.TransmissionPowerLevel, w.boostCost)
			} else {
				energyConsumption[msg.SensorID] += w.calculateEnergy(msg.NumBytes, msg.TransmissionPowerLevel, 1)
			}

		case <-w.stopChan:
			break Loop
		}

	}

	// Write to JSON
	energyStats := EnergyStats{}
	var totalEnergy float64
	for sensorID, energyCons := range energyConsumption {
		totalEnergy += energyCons
		s := SensorEnergyStats{SensorID: sensorID, EnergyConsumption: energyCons, Retries: numRetries[sensorID]}
		energyStats.Sensors = append(energyStats.Sensors, s)
	}

	var totalRetries int
	for _, retries := range numRetries {
		totalRetries += (retries - 1)
	}

	energyStats.TotalRetries = totalRetries
	energyStats.TotalEnergyConsumption = totalEnergy
	energyStats.TotalBoostMessages = boostMessages

	//fmt.Printf("Retry cost: %.2f\n", retrycost)
	//fmt.Printf("Messages sent: %d\n", messagesSent)
	//fmt.Printf("Avg. transmission level: %.2f\n", avgTransmLvl)
	fmt.Printf("Total energy consumption: %8.5f mWs\n", totalEnergy)
	fmt.Printf("Total amount of retries: %d\n", totalRetries)
	fmt.Printf("Boost messages: %d\n", boostMessages)

	// Send results to simulator
	w.energyChan <- energyStats
}

func (w *Watchdog) calculateEnergy(numBytes, step int, cost float64) float64 {
	//energyCons := float64(numBytes) * w.transmissionPowermW[step] * w.energyConsumption
	energyCons := (float64(numBytes) * 8 * w.bitCost * cost * w.transmissionPowermW[step]) / w.dataRate
	return energyCons
}

// WatchdogJSON is a struct used for printing energy related stats to file.
type WatchdogJSON struct {
	TotalRetries           int            `json:"total_retries"`
	TotalEnergyConsumption float64        `json:"total_energy_consumption"`
	Sensors                []wdSensorJSON `json:"sensors"`
}

type wdSensorJSON struct {
	SensorID          int     `json:"id"`
	EnergyConsumption float64 `json:"energy_consumption"`
	Retries           int     `json:"retries"`
}
