package utils

// MovingWindowFloat is a custom datatype for keeping track of data for a certain amount of messages at a time, with float values.
type MovingWindowFloat struct {
	capacity     int
	data         []float64
	currentIndex int
	lastID       int
}

// Total returns the current sum of the window.
func (mw *MovingWindowFloat) Total() float64 {
	return SumFloat(mw.data)
}

// Insert increments the moving window value for a given ID.
func (mw *MovingWindowFloat) Insert(id int) {
	if id == mw.lastID {
		mw.data[mw.currentIndex]++
	} else {
		nextIndex := (mw.currentIndex + 1) % mw.capacity
		mw.data[nextIndex] = 1
		mw.currentIndex = nextIndex
		mw.lastID = id
	}
}

// GetData returns a copy of the current data slice.
func (mw *MovingWindowFloat) GetData() []float64 {
	return mw.data
}

// SetValue lets you set a value directly instead of incrementing.
func (mw *MovingWindowFloat) SetValue(index int, value float64) {
	mw.data[index] = value
}

// Move moves the window one step.
func (mw *MovingWindowFloat) Move(id int) {
	if id != mw.lastID {
		nextIndex := (mw.currentIndex + 1) % mw.capacity
		mw.data[nextIndex] = 0
		mw.currentIndex = nextIndex
		mw.lastID = id
	}
}

// NewMovingWindowFloat returns a new moving window with a capacity and an initial ID, with the initial value set. Float type.
func NewMovingWindowFloat(id, capacity int, initialValue float64) *MovingWindowFloat {

	d := make([]float64, capacity)
	d[0] = initialValue

	return &MovingWindowFloat{
		capacity:     capacity,
		data:         d,
		currentIndex: 0,
		lastID:       id,
	}
}

// MovingWindowInt is a custom datatype for keeping track of data for a certain amount of messages at a time of type Int.
type MovingWindowInt struct {
	capacity     int
	data         []int
	currentIndex int
	lastID       int
}

// TotalCount returns the total count of the window.
func (mw *MovingWindowInt) TotalCount() int {
	return SumInt(mw.data)
}

// Insert increments the moving window value for a given ID.
func (mw *MovingWindowInt) Insert(id int) {

	if id == mw.lastID {
		mw.data[mw.currentIndex]++
	} else {
		nextIndex := (mw.currentIndex + 1) % mw.capacity
		mw.data[nextIndex] = 1
		mw.currentIndex = nextIndex
		mw.lastID = id
	}
}

// Move moves the window one step.
func (mw *MovingWindowInt) Move(id int) {
	if id != mw.lastID {
		nextIndex := (mw.currentIndex + 1) % mw.capacity
		mw.data[nextIndex] = 0
		mw.currentIndex = nextIndex
		mw.lastID = id
	}
}

// NewMovingWindowInt returns a new moving window with a capacity and an initial ID of type int.
func NewMovingWindowInt(id, capacity, initialValue int) *MovingWindowInt {

	d := make([]int, capacity)
	d[0] = initialValue

	return &MovingWindowInt{
		capacity:     capacity,
		data:         d,
		currentIndex: 0,
		lastID:       id,
	}
}
